#
# Copyright (c) 2020, 2021  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Implement the version comparison function."""

import functools
import re

from typing import List, Optional


RE_COMPONENT = re.compile(
    r""" ^
    (?P<num> [0-9]+ )? (?P<rest> .+ )?
$ """,
    re.X,
)


def cmp_strings(first: Optional[str], second: Optional[str]) -> int:
    """Compare two strings, possibly undefined."""
    if first is not None:
        if second is not None:
            if first == second:
                return 0
            if first > second:
                return 1
            return -1

        return 1

    if second is not None:
        return -1

    return 0


def cmp_components(first: str, second: str) -> int:
    """Compare a single pair of version components.

    - alpha4 < beta
    - alpha4 < 1
    - 1 < 1sp2
    - 1 < 2"""
    a_m = RE_COMPONENT.match(first)
    assert a_m
    a_num = a_m.group("num")
    a_rest = a_m.group("rest")

    b_m = RE_COMPONENT.match(second)
    assert b_m
    b_num = b_m.group("num")
    b_rest = b_m.group("rest")

    if a_num is not None:
        if b_num is not None:
            if a_num != b_num:
                return int(a_num) - int(b_num)
            return cmp_strings(a_rest, b_rest)

        return 1

    if b_num is not None:
        return -1

    return cmp_strings(a_rest, b_rest)


def cmp_examine_extra(a_first: str) -> int:
    """Does the version component start with a digit?

    In other words, is it an alpha/beta version or not?"""
    a_m = RE_COMPONENT.match(a_first)
    assert a_m
    if a_m.group("num") is not None:
        return 1
    return -1


def cmp_extra(first: List[str], second: List[str]) -> int:
    """Check for alpha versions in the longer list's additional components.

    Does one of the lists:
    - have more components than the other, and
    - the first extra component starts with a digit, not a letter?
    (if it starts with a letter, it's an alpha/beta/pre-release version)"""
    a_len = len(first)
    b_len = len(second)
    if a_len == b_len:
        return 0

    if a_len > b_len:
        a_first = first[b_len]
        return cmp_examine_extra(a_first)

    b_first = second[a_len]
    return -cmp_examine_extra(b_first)


def cmp_dot_components(first: str, second: str) -> int:
    """Split versions number no dots, compare them component by component."""
    a_dot = first.split(".")
    b_dot = second.split(".")
    for a_comp, b_comp in zip(a_dot, b_dot):
        res = cmp_components(a_comp, b_comp)
        if res != 0:
            return res

    # Does one of them have more dot-delimited components than the other?
    return cmp_extra(a_dot, b_dot)


# Split a version number on dashes, compare them component by component
def compare(first: str, second: str) -> int:
    """Compare two full versions strings."""
    a_dash = first.split("-")
    b_dash = second.split("-")
    for a_comp, b_comp in zip(a_dash, b_dash):
        res = cmp_dot_components(a_comp, b_comp)
        if res != 0:
            return res

    # Does one of them have more dash-delimited components than the other?
    # (don't check for beta versions in this case)
    return len(a_dash) - len(b_dash)


key_compare = functools.cmp_to_key(compare)  # pylint: disable=invalid-name

__all__ = ("compare", "key_compare")



#
# Copyright (c) 2021  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Test the command-line invocation routines."""

import sys
import unittest

from typing import List

import ddt  # type: ignore
import mock

from trivver import __main__ as tmain

from . import data


@ddt.ddt
class TestMain(unittest.TestCase):
    """Test the functions implementing the commands."""

    # pylint: disable=no-self-use

    @ddt.data(*data.VERSIONS)
    def test_compare(self, case: data.VersionSet) -> None:
        """Test the 'trivver compare left right' subcommand."""
        res: List[str] = []
        with mock.patch("builtins.print", new=res.append):
            tmain.cmd_compare(
                tmain.Config(left=case.left, right=case.right, rel="")
            )

        assert res == [data.EXPECTED_TO_STR[case.expected]]

    @ddt.data(*data.VERSIONS)
    def test_main_compare(self, case: data.VersionSet) -> None:
        """Test the 'trivver compare left right' invocation."""
        res: List[str] = []
        with mock.patch.object(
            sys, "argv", new=["trivver", "compare", case.left, case.right]
        ), mock.patch("builtins.print", new=res.append):
            tmain.main()

        assert res == [data.EXPECTED_TO_STR[case.expected]]

    @ddt.data(*data.VERSIONS)
    def test_verify(self, case: data.VersionSet) -> None:
        """Test the 'trivver verify left op right' subcommand."""
        rels = data.EXPECTED_TO_REL[case.expected]
        res: List[int] = []

        assert len(rels[False]) == 6
        for rel in rels[False]:
            with mock.patch("sys.exit", new=res.append):
                tmain.cmd_verify(
                    tmain.Config(left=case.left, right=case.right, rel=rel)
                )

        assert len(rels[True]) == 6
        for rel in rels[True]:
            with mock.patch("sys.exit", new=res.append):
                tmain.cmd_verify(
                    tmain.Config(left=case.left, right=case.right, rel=rel)
                )

        assert res == ([1] * 6) + ([0] * 6)

    @ddt.data(*data.VERSIONS)
    def test_main_verify(self, case: data.VersionSet) -> None:
        """Test the 'trivver verify left op right' subcommand."""
        rels = data.EXPECTED_TO_REL[case.expected]
        res: List[int] = []

        assert len(rels[False]) == 6
        for rel in rels[False]:
            with mock.patch.object(
                sys,
                "argv",
                new=["trivver", "verify", case.left, rel, case.right],
            ), mock.patch("sys.exit", new=res.append):
                tmain.main()

        assert len(rels[True]) == 6
        for rel in rels[True]:
            with mock.patch.object(
                sys,
                "argv",
                new=["trivver", "verify", case.left, rel, case.right],
            ), mock.patch("sys.exit", new=res.append):
                tmain.main()

        assert res == ([1] * 6) + ([0] * 6)

    def test_sort(self) -> None:
        """Test the 'trivver sort' functionality."""
        res: List[str] = []
        with mock.patch.object(
            sys.stdin,
            "readlines",
            new=lambda: [item + "\n" for item in data.UNSORTED],
        ), mock.patch("builtins.print", new=res.append):
            tmain.cmd_sort(tmain.Config(left="", right="", rel=""))

        assert res == ["\n".join(data.SORTED)]

    def test_main_sort(self) -> None:
        """Test the 'trivver sort' functionality."""
        res: List[str] = []
        with mock.patch.object(
            sys, "argv", new=["trivver", "sort"]
        ), mock.patch.object(
            sys.stdin,
            "readlines",
            new=lambda: [item + "\n" for item in data.UNSORTED],
        ), mock.patch(
            "builtins.print", new=res.append
        ):
            tmain.main()

        assert res == ["\n".join(data.SORTED)]



#
# Copyright (c) 2020, 2021  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Test the version comparison routines."""

import unittest

import ddt  # type: ignore

import trivver

from . import data


def sign(value: int) -> int:
    """Return the sign of an integer."""
    if value > 0:
        return 1
    if value < 0:
        return -1
    return 0


@ddt.ddt
class TestVersions(unittest.TestCase):
    """Test the version comparison routines."""

    # pylint: disable=no-self-use

    @ddt.data(*data.VERSIONS)
    @ddt.unpack
    def test_compare(self, left: str, right: str, expected: int) -> None:
        """Compare a single pair of versions."""
        assert sign(trivver.compare(left, right)) == sign(expected)
        assert sign(trivver.compare(right, left)) == sign(-expected)

    def test_sort(self) -> None:
        """Sort a list of versions."""
        assert sorted(data.UNSORTED, key=trivver.key_compare) == data.SORTED



#
# Copyright (c) 2020, 2021  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Definitions and data for the trivver unit tests."""

from typing import NamedTuple


class VersionSet(NamedTuple):
    """A basic test case for comparing versions."""

    left: str
    right: str
    expected: int


VERSIONS = [
    VersionSet(left="1.0", right="2.0", expected=-1),
    VersionSet(left="1.0", right="1.0.1", expected=-1),
    VersionSet(left="1.0", right="1.0a", expected=-1),
    VersionSet(left="1.0", right="1.0", expected=0),
    VersionSet(left="1.0a", right="1.0a", expected=0),
    VersionSet(left="0.1.0.b", right="0.1.0", expected=-1),
    VersionSet(
        left="3.10.0-1062.1.1.el7.x86_64",
        right="3.10.0-983.13.1.el7.x86_64",
        expected=1,
    ),
    VersionSet(
        left="3.10.0-1062.1.1.el7.x86_64",
        right="3.10.0-1062.1.1.el6.x86_64",
        expected=1,
    ),
]

EXPECTED_TO_STR = {
    -1: "<",
    0: "=",
    1: ">",
}

EXPECTED_TO_REL = {
    -1: {
        False: ["=", ">", ">=", "eq", "gt", "ge"],
        True: ["<", "<=", "!=", "lt", "le", "ne"],
    },
    0: {
        False: ["<", ">", "!=", "lt", "gt", "ne"],
        True: ["=", "<=", ">=", "eq", "le", "ge"],
    },
    1: {
        False: ["<", "=", "<=", "lt", "eq", "le"],
        True: [">", ">=", "!=", "gt", "ge", "ne"],
    },
}

UNSORTED = [
    "3.10.0-1062.1.1.el7.x86_64",
    "1.0a",
    "0.1.0.b",
    "1.0",
    "3.10.0-983.13.1.el7.x86_64",
    "0.1.0",
    "1.0.1",
    "2.0",
    "3.10.0-1062.1.1.el6.x86_64",
]
SORTED = [
    "0.1.0.b",
    "0.1.0",
    "1.0",
    "1.0.1",
    "1.0a",
    "2.0",
    "3.10.0-983.13.1.el7.x86_64",
    "3.10.0-1062.1.1.el6.x86_64",
    "3.10.0-1062.1.1.el7.x86_64",
]



#
# Copyright (c) 2021  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Command-line version comparison tool."""

import argparse
import sys

from typing import Callable, Dict, NamedTuple, Tuple

from . import trivver


V_HANDLERS: Dict[str, Callable[[int], bool]] = {
    "<": lambda res: res < 0,
    "=": lambda res: res == 0,
    ">": lambda res: res > 0,
    "<=": lambda res: res <= 0,
    ">=": lambda res: res >= 0,
    "!=": lambda res: res != 0,
}
V_ALIASES = {
    "lt": "<",
    "eq": "=",
    "gt": ">",
    "le": "<=",
    "ge": ">=",
    "ne": "!=",
}
V_CHOICES = sorted(V_HANDLERS) + sorted(V_ALIASES)


class Config(NamedTuple):
    """Runtime configuration for the version comparison tool."""

    left: str
    right: str
    rel: str


def cmd_compare(cfg: Config) -> None:
    """Compare the two versions specified."""
    res = trivver.compare(cfg.left, cfg.right)
    if res < 0:
        print("<")
    elif res > 0:
        print(">")
    else:
        print("=")


def cmd_verify(cfg: Config) -> None:
    """Verify that a relation holds true for the specified versions."""

    rel = V_ALIASES.get(cfg.rel, cfg.rel)
    handler = V_HANDLERS.get(rel)
    if handler is None:
        sys.exit(
            (
                "Invalid relation '{rel}' specified, "
                "must be one of {choices}"
            ).format(rel=cfg.rel, choices=" ".join(V_CHOICES))
        )

    res = trivver.compare(cfg.left, cfg.right)
    sys.exit(0 if handler(res) else 1)


def cmd_sort(_cfg: Config) -> None:
    """Sort a list of versions supplied on the standard input stream."""
    lines = [line.strip() for line in sys.stdin.readlines()]
    print("\n".join(sorted(lines, key=trivver.key_compare)))


def parse_args() -> Tuple[Config, Callable[[Config], None]]:
    """Parse the command-line arguments."""
    parser = argparse.ArgumentParser(prog="trivver")

    subp = parser.add_subparsers(help="trivver subcommands")

    cmd_p = subp.add_parser("compare", help="compare the specified versions")
    cmd_p.add_argument("left", type=str, help="the first version string")
    cmd_p.add_argument("right", type=str, help="the second version string")
    cmd_p.set_defaults(func=cmd_compare)

    cmd_p = subp.add_parser("verify", help="verify that a relation holds true")
    cmd_p.add_argument("left", type=str, help="the first version string")
    cmd_p.add_argument(
        "rel", type=str, choices=V_CHOICES, help="the relation to verify"
    )
    cmd_p.add_argument("right", type=str, help="the second version string")
    cmd_p.set_defaults(func=cmd_verify)

    cmd_p = subp.add_parser(
        "sort", help="sort a list of versions from stdin to stdout"
    )
    cmd_p.set_defaults(func=cmd_sort)

    args = parser.parse_args()

    func = getattr(args, "func", None)
    if func is None:
        sys.exit("No command specified")

    return (
        Config(
            left=getattr(args, "left", ""),
            right=getattr(args, "right", ""),
            rel=getattr(args, "rel", ""),
        ),
        func,
    )


def main() -> None:
    """Parse the command-line arguments, perform the required actions."""
    cfg, func = parse_args()
    func(cfg)


if __name__ == "__main__":
    main()
